/******************************************************************************
		Taylor Polynomial Approximation - Algorithm 1.1B
*******************************************************************************

To approximate the 4th Taylor Polynomial, P4(x), from the function f(x) about
the point x0.  Note: f(5)(x) must exist and x0 must not be at a discontinuity.

INPUT the point x0; evaluation points x; the function f():

OUTPUT approximation to the 4th Taylor Polynomial.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include <math.h>		/* for math functions */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097


main()
{
  double x0, x, h=0.001, *U, *Df, f(), tmp, approx;
  int i, ch;

  get_title();

  printf("Taylor Polynomial Approximation - Algorithm 1.1B\n\n");
  fprintf(stderr, "Taylor Polynomial Approximation - Algorithm 1.1B\n\n");

  printf("f(x) = exp(-x*x)\n\n");
  /* Needs updating  $  */

  fprintf(stderr, "Enter point x0: ");
  scanf("%lf", &x0);
  printf("Point x0 = %lg\n\n", x0);

  printf("GENERAL FORM:\n");
  printf("P4(x) = f(x0)\n\t+ f'(x0)(x-x0)\n\t+ (f''(x0)/2!)(x-x0)^2\n\t");
  printf("+ (f'''(x0)/3!)(x-x0)^3\n\t+ (f(4)(x0)/4!)(x-x0)^4\n\t");  
  printf("+ R4(x)\n\n");

  U = dvector(-2,2);		/* Stores values of f(x) to be used. */
  Df = dvector(0,4);		/* Stores derivatives of f(x) to be used. */

  for (i= -2;i<=2;i++)
    U[i] = f(x0 + i*h);

  Df[0] = U[0];			/* Compute derivatives. */
  Df[1] = (-U[2] + 8*U[1] - 8*U[-1] + U[-2]) / (12*h);
  Df[2] = (-U[2] + 16*U[1] - 30*U[0] + 16*U[-1] - U[-2]) / (12*h*h);
  Df[3] = (U[2] - 2*U[1] + 2*U[-1] - U[-2]) / (2*h*h*h);
  Df[4] = (U[2] - 4*U[1] + 6*U[0] - 4*U[-1] + U[-2]) / pow(h,4.0);

  Df[2] /= 2.0;			/* Divide by factorials. */
  Df[3] /= 6.0;
  Df[4] /= 24.0;

  fprintf(stderr, "P4(x) = (%lg)\n\t+ (%lg)(x-%lg)\n\t", Df[0], Df[1], x0);
  fprintf(stderr,
	"+ (%lg)(x-%lg)^2\n\t+ (%lg)(x-%lg)^3\n\t", Df[2], x0, Df[3], x0);
  fprintf(stderr, "+ (%lg)(x-%lg)^4\n\t+ R4(x)\n", Df[4], x0);

  printf("P4(x) = (%lg)\n\t+ (%lg)(x-%lg)\n\t", Df[0], Df[1], x0);
  printf("+ (%lg)(x-%lg)^2\n\t+ (%lg)(x-%lg)^3\n\t", Df[2], x0, Df[3], x0);
  printf("+ (%lg)(x-%lg)^4\n\t+ R4(x)\n\n", Df[4], x0);

  do {

    fprintf(stderr, "\nEvaluate P4(x) at x = ?: ");
    scanf("%lf", &x);

    tmp = x - x0;
    approx = (((((Df[4] * tmp) + Df[3]) * tmp + Df[2])
		* tmp + Df[1]) * tmp + Df[0]);
    fprintf(stderr, "\nP4(%lg) = %lg", x, approx);
    printf("P4(%lg) = %lg\n", x, approx);

    fprintf(stderr,
	"\nWould you like to evaluate P4(x) at another point? (y/n): ");
    ch = getchar();
    ch = tolower(getchar());

  } while (ch == 'y');

}

/*****************************************************************************/
double f(x)		/* Function to evaluate, f(x). */
double x;
{
  return (exp(-x*x));
}
/*****************************************************************************/
