/******************************************************************************
		Clamped Cubic Spline - Algorithm 3.5
*******************************************************************************

To construct the cubic spline interpolant S for the function f(), defined at
the numbers x0 < x1 < ... < xn, satisfying S'(x0) = f'(x0) and S'(xn) = f'(xn):

INPUT n; x0,x1,...,xn; either generate a(i) = f(x(i)) for i = 0,1,...,n or
input a(i) for i = 1,...,n; FPO = f'(x0); FPN = f'(xn).

OUTPUT a(j), b(j), c(j), d(j), for j = 0,1,...,n-1.
(Note: S(x) = S(j)(x) = a(j) + b(j)*(x-x(j)) + c(j)*(x-x(j))^2
	+ d(j)*(x-x(j))^3 for x(j) <= x <= x(j+1).)

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include <math.h>		/* For math functions */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097


main()
{
  double *a, *alpha, *b, *c, *d, *h, *l, *mu, *x, *z, FPO, FPN, f();
  double area, X, tmp, simpson();
  int i, j, n, ch;

  get_title();

  printf("Clamped Cubic Spline - Algorithm 3.5\n\n");
  fprintf(stderr, "Clamped Cubic Spline - Algorithm 3.5\n\n");

  fprintf(stderr, "Enter number of data points, n+1: ");
  scanf("%d", &n);
  if (n <= 0)
    naaerror("n must be > 0.");	/* ERROR. */
  printf("Uses %d data points.\n", n);
  n--;
  printf("Computes %d coefficient sets for a Clamped Cubic Spline.\n\n", n);

  printf("S(x) =  S(j)(x) = a(j) + b(j)*(x-x(j)) + c(j)*(x-x(j))^2 ");
  printf("+ d(j)*(x-x(j))^3\n\tfor x(j) <= x <= x(j+1)\n");

  a = dvector(0,n);
  alpha = dvector(0,n);
  b = dvector(0,n);
  c = dvector(0,n);
  d = dvector(0,n);
  h = dvector(0,n);
  l = dvector(0,n);
  mu = dvector(0,n);
  x = dvector(0,n);
  z = dvector(0,n);

  fprintf(stderr, "Do the a(i) = f(x(i)) need to be generated");
  fprintf(stderr, " from f(x)? (y/n): ");
  ch = getchar();		/* filters out New Line/Carrage Return */
  ch = tolower(getchar());
  if (ch == 'y')
    printf("\nf(x) = cos(PI*x)\n");
    /* Needs updating     $      */

  for (i=0;i<=n;i++) {		/* Enter x(i). */
    fprintf(stderr, "Enter x%d: ", i);
    scanf("%lf", &x[i]);
    if (ch != 'y') {		/* Input a(i). */
      fprintf(stderr, "Enter f(%lg): ", x[i]);
      scanf("%lf", &a[i]);
      fprintf(stderr, "\n");
    } else
      a[i] = f(x[i]);		/* Use function f(x(i)) for a(i). */
  }
				 /* Enter derivatives at the endpoints. */
  fprintf(stderr, "Enter FPO = f'(%lg): ", x[0]);
  scanf("%lf", &FPO);
  printf("FPO = f'(%lg) = %.9lg\n", x[0], FPO);

  fprintf(stderr, "Enter FPN = f'(%lg): ", x[n]);
  scanf("%lf", &FPN);
  printf("FPN = f'(%lg) = %.9lg\n", x[n], FPN);

  /*-------------------------------------------------------------------------*/
  printf("\n\n i\tx(i)\t a(i) = f(x)\t b(i)\t\t c(i)\t\t d(i)\n");
  printf("---------------------------------------");
  printf("----------------------------------------\n");
  /*-------------------------------------------------------------------------*/

  /* STEP #1 */
  for (i=0;i<n;i++)
    h[i] = x[i+1] - x[i];

  /* STEP #2 */
  alpha[0] = 3.0*(a[1]-a[0])/h[0] - 3.0*FPO;
  alpha[n] = 3.0*FPN - 3.0*(a[n]-a[n-1])/h[n-1];

  /* STEP #3 */
  for (i=1;i<n;i++)
    alpha[i] = (3.0*(a[i+1]*h[i-1] - a[i]*(x[i+1]-x[i-1]) + a[i-1]*h[i])) /
		 (h[i-1]*h[i]);

  /* STEP #4 */
  l[0] = 2.0*h[0];		/* Steps 4, 5, 6, and part of step 7 solve */
  mu[0] = 0.5;			/* a tridiagonal system using Algorithm 6.7. */
  z[0] = alpha[0]/l[0];

  /* STEP #5 */
  for (i=1;i<n;i++) {
    l[i] = 2.0*(x[i+1]-x[i-1]) - h[i-1]*mu[i-1];
    mu[i] = h[i]/l[i];
    z[i] = (alpha[i]-h[i-1]*z[i-1]) / l[i];
  }

  /* STEP #6 */
  l[n] = h[n-1]*(2.0-mu[n-1]);;
  z[n] = (alpha[n]-h[n-1]*z[n-1])/l[n];
  c[n] = z[n];

  /* STEP #7 */
  for (j=n-1;j>=0;j--) {
    c[j] = z[j] - mu[j]*c[j+1];
    b[j] = (a[j+1]-a[j])/h[j] - h[j]*(c[j+1]+2.0*c[j])/3.0;
    d[j] = (c[j+1]-c[j])/(3.0*h[j]);
  }

  /* STEP #8 */
  area = 0.0;
  for (i=0;i<n;i++) {
    printf(" %d\t% .3lg\t% .9lf\t% .9lf\t% .9lf\t% .9lf\n",
	 i, x[i], a[i], b[i], c[i], d[i]);
				/* Integrate the spline. */
    area += simpson(x[i], x[i+1], a[i], b[i], c[i], d[i], a[i+1]);
  }
				/* Print a(i) only in last row. */
  printf(" %d\t% .2lg\t% .9lf\n", n, x[n], a[n]);

  printf("\n\n");		/* Print integral. */
  printf("    [%lg\n", x[n]);
  printf("I = |  S(x) dx = %.11lg\n", area);
  printf("    ]%lg\n\n", x[0]);

  fprintf(stderr, "Would you like to evaluate f(x), f'(x), and f''(x)");
  fprintf(stderr, " at a certain x? (y/n): ");
  ch = getchar();		/* filters out New Line/Carrage Return */
  ch = tolower(getchar());

  if (ch == 'y') {

    fprintf(stderr, "\nEnter a value of x between %lg and %lg\n", x[0], x[n]);
    fprintf(stderr, "for which f(x), f'(x), and f''(x) may be evaluated.\n\n");
    scanf("%lf", &X);

    for (j=0;j<n;j++) {		/* Search for correct x interval. */
      if ((X >= x[j]) && (X < x[j+1]))
        break;			/* keep j. */
    }

    tmp = a[j] + b[j]*(X-x[j]) + c[j]*pow(X-x[j], 2.0) + d[j]*pow(X-x[j], 3.0);
    printf("f(%lg) = S(%lg) = %.11lg\n", X, X, tmp);	/* Find f(x). */

    tmp = b[j] + 2.0*c[j]*(X-x[j]) + 3.0*d[j]*pow(X-x[j], 2.0);
    printf("f'(%lg) = S'(%lg) = %.11lg\n", X, X, tmp);	/* Find f'(x). */

    tmp = 2.0*c[j] + 6.0*d[j]*(X-x[j]);
    printf("f''(%lg) = S''(%lg) = %.11lg\n", X, X, tmp); /* Find f''(x). */

  }

}  /* STOP */

/******************************************************************************
	Composite Simpson's Rule - Algorithm 4.1 (with f() = S())
******************************************************************************/
double simpson(aa,bb,a,b,c,d,a2)
double aa, bb;		/* Bounds of integration. aa = x(j), bb = x(j+1) */
double a,b,c,d;		/* Cubic polynomial coefficients. */
double a2;		/* a = f(aa), a2 = f(bb) */
{
  double h, x, X, XI, XI0, XI1, XI2, f();
  int i, n=10;			/* n = 2m (even) for subintervals. */

  h = (bb-aa)/n;
  XI0 = a + a2;			/* S(xj) = f(xj) = a(j). */
  XI1 = 0.0;			/* Summation of S(x(2i-1)). */
  XI2 = 0.0;			/* Summation of S(x(2i)). */

  for (i=1;i<n;i++) {
    X = aa + i*h;
    x = X - aa;			/* x = (x-xj) = (x-aa) */
    if (i % 2 == 0)
      XI2 += ((d*x+c)*x+b)*x+a;	/* Compute S(x) for i even. */
    else
      XI1 += ((d*x+c)*x+b)*x+a;	/* Compute S(x) for i odd. */
  }

  XI = h*(XI0 + 2.0*XI2 + 4.0*XI1) / 3.0;
  return (XI);

}
/* S(x) = S(j)(x) = a(j) + b(j)*(x-x(j)) + c(j)*(x-x(j))^2 + d(j)*(x-x(j))^3 */

/*****************************************************************************/
/*****************************************************************************/
double f(x)		/* Function to evaluate, f(x).  May be unnecessary. */
double x;
{
  return (cos(PI*x));
}
/*****************************************************************************/
