/******************************************************************************
	Midpoint, Modified Euler, and Heun's Methods - Algorithm 5.1B
*******************************************************************************

To approximate the solution of the initial-value problem
	y' = f(t,y),  a <= t <= b,  y(a) = alpha,
at (N+1) equally spaced numbers in the interval [a,b]:

INPUT endpoints a, b; integer N or interval h; initial condition alpha; the
function f(); which method(s) to try:

OUTPUT approximation w to y at the (N+1) values of t.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include <math.h>		/* for math functions */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097

double a, b, h=0.0, alpha, f();
int N;


main()
{
  int ch, OK=0, type;

  get_title();

  printf("Midpoint, Modified Euler, and Heun's Methods - Algorithm 5.1B\n\n");
  fprintf(stderr,
	"Midpoint, Modified Euler, and Heun's Methods - Algorithm 5.1B\n\n");

  printf("y' = f(t,y) = -y + t*t + 1\n\n");
  /* Needs updating         $          */

  fprintf(stderr, "Enter endpoint a: ");
  scanf("%lf", &a);
  fprintf(stderr, "Enter endpoint b: ");
  scanf("%lf", &b);
  printf("Solving for points on interval [%lg,%lg].\n", a, b);

  fprintf(stderr, "Enter initial condition y(%lg) = alpha: ", a);
  scanf("%lf", &alpha);
  printf("y(%lg) = alpha = %lg\n", a, alpha);

  do {

    fprintf(stderr,"\nSelect desired method:\n");
    fprintf(stderr,"    (1)   Midpoint Method\n");
    fprintf(stderr,"    (2)   Modified Euler Method\n");
    fprintf(stderr,"    (3)   Heun's Method\n");
    fprintf(stderr,"    (4)   All of the above (1-3).\n");
    fprintf(stderr, "\n");
    scanf("%d", &type);

    if ((type < 1) || (type > 4))
      naaerror("Invalid selection.  Use (1-4 only).");

    do {
      fprintf(stderr, "Do you want to enter N or h? ");
      ch = getchar();		/* Filters out Carriage Return/Linefeed. */
      ch = tolower(getchar());
      if (ch == 'n') {
        fprintf(stderr, "Enter number of subintervals N: ");
        scanf("%d", &N);
        h = (b-a)/N;
        OK = 1;
      }
      if (ch == 'h') {
        fprintf(stderr, "Enter interval number h: ");
        scanf("%lf", &h);
        N = (int) ceil((b-a)/h);
        OK = 1;
      }
      if (OK==0)
        fprintf(stderr, "\tTRY AGAIN.\n");
    } while (OK==0);

    printf("Number of subintervals N = %d\n", N);
    printf("Interval number h = %lg\n", h);

    Evaluate(type);		/* Find solution using Method "type". */

    if (type == 4)
      exit (1);
    else {
      fprintf(stderr, "\n\nI'm all done!\n");
      fprintf(stderr, "Would you like to try another method? (y/n): ");
      ch = getchar();
      ch = tolower(getchar());
    }
  } while (ch == 'y');

}

/*****************************************************************************/
Evaluate(type)		/* Evaluates a specific derivative formula. */
int type;
{
  double Midpoint(), Modified_Euler(), Heuns();
  double t, w, w1, w2, w3;
  int i;

  switch (type) {
    case 1: printf("Using Midpoint Method:\n");
            printf("\n t\t  w = y\n-----------------------\n");
            break;
    case 2: printf("Using Modified Euler Method:\n");
            printf("\n t\t  w = y\n-----------------------\n");
            break;
    case 3: printf("Using Heun's Method:\n");
            printf("\n t\t  w = y\n-----------------------\n");
            break;
    case 4: printf("\n t\t  Midpoint\t  Modified Euler  Heun's\n");
            printf("---------------------------");
            printf("----------------------------\n");
            break;
  }

  t = a;

  if (type != 4) {
    w = alpha;
    printf("% .2lf\t% .11lf\n", t, w);
    for (i=1;i<=N;i++) {
      switch (type) {		/* compute wi */
        case 1:  w = Midpoint(t,w);  break;
        case 2:  w = Modified_Euler(t,w);  break;
        case 3:  w = Heuns(t,w);  break;
      }
      t += h;			/* compute ti */
      printf("% .2lf\t% .11lf\n", t, w);
    }
    printf("\n");

  } else {
    w1 = alpha; w2 = alpha; w3 = alpha;
    printf("% .2lf\t% .9lf\t% .9lf\t% .9lf\n", t, w1, w2, w3);
    for (i=1;i<=N;i++) {	/* compute wi */
      w1 = Midpoint(t,w1);
      w2 = Modified_Euler(t,w2);
      w3 = Heuns(t,w3);
      t += h;			/* compute ti */
      printf("% .2lf\t% .9lf\t% .9lf\t% .9lf\n", t, w1, w2, w3);
    }
  }
}

/*****************************************************************************/
double Midpoint(t,w)		/* Midpoint Method */
double t, w;
{
  return (w + h*f(t+(h/2), w+(h/2)*f(t,w)));
}

/*****************************************************************************/
double Modified_Euler(t,w)	/* Modified Euler Method (RK2) */
double t, w;
{
  double F;
  F = f(t,w);

  return (w + (h/2)*(F + f(t+h, w+h*F)));
}

/*****************************************************************************/
double Heuns(t,w)		/* Heun's Method (RK2) */
double t, w;
{
  double F;
  F = f(t,w);

  return (w + (h/4)*(F + 3*f(t+(2*h/3), w+(2*h/3)*F)) );
}

/*****************************************************************************/
/*****************************************************************************/
double f(t,y)		/* Function y' = f(t,y). */
double t, y;
{
  return (-y + t*t + 1);
}
/*****************************************************************************/
