/******************************************************************************
		Fast Fourier Transformation - Algorithm 8.1
*******************************************************************************

To compute the discrete approximation
	F(x) = (1/m) * Summation of c(k)*e(ikx)  for k = 0,..., 2m-1,
	F(x) = (1/m) * Summation of c(k)*(cos(kx) + i*sin(kx))  for k = 0,..., 2m-1,
where i = (-1), for the data {(x(j), y(j))} for j = 0,..., 2m-1 where m = 2^p
and x(j) = -PI + j*PI/m for j = 0,..., 2m-1:

INPUT m, (p is determined by m); y(0), y(1),..., y(2m-1); interval a,b.

OUTPUT complex numbers c(0),..., c(2m-1); real numbers a(0),..., a(m);
b(1),..., b(m-1).  Note: the a(j) and b(j) coefficients are used for a
trigometric interpolatory polynomial S(n) on the interval [a,b].

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* For scanf() and printf() */
#include <math.h>		/* for math functions */
#include "complex.h"		/* For complex arithmetic. */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#include "naa_uti2.h"		/* Numerical Analysis Algorithms Util. II. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097


main()
{
  double *A, *B, *x, a, b, temp2, f();
  fcomplex *C, *Y, Eta, *Xi, temp, *C_carat;
  long i, j, m, p, pp, q, K, K2, L, M;
  int ch, *k;

  get_title();

  printf("Fast Fourier Transformation - Algorithm 8.1\n\n");
  fprintf(stderr, "Fast Fourier Transformation - Algorithm 8.1\n\n");

  fprintf(stderr, "Enter the degree of the polynomial, m = 2^p: ");
  scanf("%ld", &m);
  if (m <= 0)
    naaerror("m must be > 0.");	/* ERROR. */
  temp2 = log((double)(m)) / log(2.0);
  if (ceil(temp2) != floor(temp2))	/* ERROR */
    naaerror("m is not a power of 2.");
  p = (long) temp2;
  printf("m = 2^p = %ld, where p = %ld.\n\n", m, p);

  A = dvector(0,m);		/* Real A coefficients */
  B = dvector(1,m-1);		/* Real B coefficients */
  x = dvector(0,2*m-1);		/* Real x coefficients */
  C = cvector(0,2*m-1);		/* Complex C coefficients */
  C_carat = cvector(0,2*m-1);	/* Complex C_carat coefficients */
  Y = cvector(0,2*m-1);		/* Complex Y coefficients */
  Xi = cvector(0,2*m-1);	/* Complex Xi coefficients */
  k = ivector(0,2*m-1);		/* For decomposing K */

  fprintf(stderr, "Is the desired interval [-PI, PI]? (y/n): ");
  ch = getchar();		/* Filters out Carriage Return/Line Feed. */
  ch = tolower(getchar());

  if (ch == 'y') {
    printf("Using [-PI, PI] for the interval.\n");
    for (j=0;j<2*m;j++)
      x[j] = -PI + j*(PI/m);

  } else {

    fprintf(stderr, "Enter start of interval, a: ");
    scanf("%lf", &a);
    fprintf(stderr, "Enter end of interval, b: ");
    scanf("%lf", &b);
    printf("Using [%lG, %lG] for the interval.\n\n", a, b);

    printf("Transforming [%lG, %lG] to [-PI, PI] gives:\n\n", a, b);
    printf("k\t X[k]\t\t\t Z[k]\n");
    printf("--------------------------------------------\n");

    for (j=0;j<2*m;j++) {
      /* Calculate x(j) then transforms [a,b] to [-PI,PI]. */
      x[j] = a+j*((b-a)/(2.0*m));
      printf("%ld\t% .8lf\t--->\t", j, x[j]);
      printf("% .8lf\n", PI*(x[j] - (a+b)/2.0));
    }
  }

  fprintf(stderr,
	"Will you transform a function, f(x), or tabulated data? (f/d): ");
  ch = getchar();		/* Filters out Carriage Return/Line Feed. */
  ch = tolower(getchar());
  switch (ch) {
    case 'f':
    case 'y':  for (j=0;j<2*m;j++) {	/* Calculates Y values from f(x). */
                 Y[j].r = f(x[j]);
                 Y[j].i = 0.0;
               }
               break;

    case 'd':
    case 'n':  /* Output Y values (both real and imaginary). */
               printf("\n\n k\t Y(k)\n");
               printf("------------------------\n");

               fprintf(stderr, "Enter the Y values:\n");
               for (j=0;j<2*m;j++) {
                 fprintf(stderr, "X[%ld] = % lG\n", j, x[j]);

                 fprintf(stderr, "Re(Y[%ld]) = ", j);
                 scanf("%f", &Y[j].r);
                 printf("%ld\t% lG", j, Y[j].r);

                 fprintf(stderr, "Im(Y[%ld]) = ", j);
                 scanf("%f", &Y[j].i);
                 fprintf(stderr, "\n");

                 if (Y[j].i < 0.0)
                   printf(" - ");
                 else
                   printf(" + ");

                 printf("%lGi\n", fabs(Y[j].i));
               }
               break;

    default:  naaerror("Invalid choice.  Choose 'f' for f(x) or 'd' for data.");
              break;
  }

  /* STEP #1 */
  M = m;
  q = p;
/*
  Zeta.r = cos(PI/m);    Used in step #3.
  Zeta.i = sin(PI/m);
*/

  /* STEP #2 */
  for (j=0;j<2*m;j++) {
    C[j].r = Y[j].r;
    C[j].i = Y[j].i;
  }

  /* STEP #3 */
  for (j=1;j<=M;j++) {
    Xi[j].r = cos((PI*j) / (float) m);
    Xi[j].i = sin((PI*j) / (float) m);
    Xi[j+M].r = -Xi[j].r;
    Xi[j+M].i = -Xi[j].i;
  }

  /* STEP #4 */
  K = 0L;
  Xi[0].r = 1.0;
  Xi[0].i = 0.0;

  /* STEP #5 */
  for (L=1L;L<=p+1;L++) {	/* Do Steps 6-12. */

    /* STEP #6 */
    while (K < 2*m-1) {		/* Do Steps 7-11. */

      /* STEP #7 */
      for (j=1L;j<=M;j++) {	/* Do Steps 8-10. */

        /* STEP #8 */
        /* Let K = k(p)*2^p + k(p-1)*2^(p-1) + ... + k(1)*2 + k(0). */
        decompose(K,p,k);	/* Decompose K into k[]. */
        /* K1 = K / (long) (pow(2.0, (double) q)); */	/* Never used. */
        K2 = 0L;
        pp = p;
        for (i=q;i<=p;i++)
          K2 += (long) (k[i] * ldexp(1.0, (int) pp--));

        /* STEP #9 */
        Eta = Cmul(C[K+M], Xi[K2]);
        C[K+M] = Csub(C[K], Eta);
        C[K] = Cadd(C[K], Eta);

        /* STEP #10 */
        K++;
      }

      /* STEP #11 */
      K += M;
    }

    /* STEP #12 */
    K = 0L;
    M /= 2L;
    q--;
  }

  /* STEP #13 */
  while (K < 2*m-1) {		/* Do Steps 14-16. */

    /* STEP #14 */
    /* Let K = k(p)*2^p + k(p-1)*2^(p-1) + ... + k(1)*2 + k(0). */
    decompose(K,p,k);		/* Decompose K into k[]. */
    j = 0L;
    pp = p;
    for (i=0L;i<=p;i++)
      j += (long) (k[i] * ldexp(1.0, (int) pp--));

    /* STEP #15 */
    if (j > K) {		/* Interchange C[j] and C[k]. */
      temp.r = C[j].r;
      temp.i = C[j].i;
      C[j].r = C[K].r;
      C[j].i = C[K].i;
      C[K].r = temp.r;
      C[K].i = temp.i;
    }

    /* STEP #16 */
    K++;
  }

  /* STEP #17 */
  A[0] = C[0].r / m;
  temp.r = cos(m * PI);
  temp.i = -sin(m * PI);
  temp = Cmul(temp, C[m]);
  A[m] = temp.r / m;

  C_carat[0].r = A[0];
  C_carat[0].i = 0.0;
  C_carat[m].r = A[m];
  C_carat[m].i = 0.0;

  /* STEP #18 */
  for (j=1;j<m;j++) {
    temp.r = cos(j*PI);
    temp.i = -sin(j*PI);
    temp = Cmul(temp, C[j]);
    A[j] = temp.r / m;
    B[j] = temp.i / m;

    C_carat[j].r = A[j];
    C_carat[j].i = B[j];
  }

  /* STEP #19 */
  /* Output complex terms. */
  printf("\n\n k\t Re(C)\t\t Im(C)\t\t Re(C_carat)\t Im(C_carat)\n");
  printf("------");
  for (i=1;i<=4;i++)
    printf("----------------");
  printf("\n");

  for (j=0L;j<2*m;j++) {
    printf(" %ld\t% 9.9G\t% 9.9G", j, C[j].r, C[j].i);
    if (j <= m)
      printf("\t% 9.9G\t% 9.9G", C_carat[j].r, C_carat[j].i);
    if (j > m)
      printf("\t% 9.9G\t% 9.9G", C_carat[m-(j%m)].r, -C_carat[m-(j%m)].i);
    printf("\n");
  }

  /* Output polynomial coefficients. */
  printf("\n\n k\t A = Re(C_carat) B = Im(C_carat)\n");
  printf("------------------------------------------\n");

  for (j=0L;j<=m;j++) {
    printf(" %ld\t% 9.9lG", j, A[j]);
    if (j >= 1 && j < m)
      printf("\t% 9.9lG", B[j]);
    printf("\n");
  }

  printf("\nUse the above coefficients in the trigonometric interpolatory\n");
  printf("polynomial S(x) of degree %ld:\n\n", m);
  printf("S(x) = A[0] / 2  +  A[%ld] * cos(%ldx)  +  ", m, m);
  printf("Summation of  (A[k] * cos(kx)  +  B[k] * sin(kx))\n");
  printf("       summed from k = 1 to %ld.\n\n", m-1);

}  /* STOP */

/*****************************************************************************/
decompose(K,max,k)
long K, max;
int *k;
{
  long i, temp;

  for (i=max;i>=0L;i--) {		/* Decompose K from left to right. */
    temp = (long) ldexp(1.0, (int) i);
    if (K >= temp) {
      K -= temp;
      k[i] = 1;
    } else
      k[i] = 0;
  }
}

/*****************************************************************************/
/*****************************************************************************/
double f(x)		/* Function to evaluate, f(x).  May not be needed. */
double x;
{
  return(exp(-x));
}
/*****************************************************************************/
