/******************************************************************************
			Linear Shooting - Algorithm 11.1
*******************************************************************************

To approximate the solution of the linear boundry-value problem

	-y'' + p(x)y' + q(x)y + r(x) = 0,  a <= x <= b,  y(a) = alpha,  y(b) = beta:

(Note: Equations (11.5) and (11.6) are written as first-order systems and
solved.)

INPUT endpoints a, b; boundry conditions alpha, beta; number of subintervals N or step
size h; the functions p(), q(), and r():

OUTPUT approximations w1,i to y(xi); w2,i to y'(xi) for each i = 0,1,...,N.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include <math.h>		/* for math functions */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097


main()
{
  double **k, **K, **u, **v, p(), q(), r();
  double a, b, alpha, beta, h=0.0, x, w10, w20, W1, W2;
  int i, N, OK=0, ch;

  get_title();

  printf("Linear Shooting - Algorithm 11.1\n\n");
  fprintf(stderr, "Linear Shooting - Algorithm 11.1\n\n");

  printf("y'' = p(x)y' + q(x)y + r(x)\n");
  printf("p(x) = -2.0/x\n");
  printf("q(x) = 2.0/(x*x)\n");
  printf("r(x) = sin(log(x))/(x*x)\n\n");
  /* Needs updating    $    $    $     */

  fprintf(stderr, "Enter endpoint a: ");
  scanf("%lf", &a);
  fprintf(stderr, "Enter endpoint b: ");
  scanf("%lf", &b);
  printf("Solving for points on interval [%lg,%lg].\n", a, b);

  fprintf(stderr, "Enter boundry condition y(%lg) = alpha: ", a);
  scanf("%lf", &alpha);
  printf("Boundry condition y(%lg) = alpha = %lg\n", a, alpha);

  fprintf(stderr, "Enter boundry condition y(%lg) = beta: ", b);
  scanf("%lf", &beta);
  printf("Boundry condition y(%lg) = beta = %lg\n", b, beta);

  do {
    fprintf(stderr, "Do you want to enter N or h? ");
    ch = getchar();		/* Filters out Carriage Return/Linefeed. */
    ch = tolower(getchar());
    if (ch == 'n') {
      fprintf(stderr, "Enter number of subintervals N: ");
      scanf("%d", &N);
      h = (b-a)/N;
      OK = 1;
    }
    if (ch == 'h') {
      fprintf(stderr, "Enter interval number h: ");
      scanf("%lf", &h);
      N = (int) ceil((b-a)/h);
      OK = 1;
    }
    if (OK==0)
      fprintf(stderr, "\tTRY AGAIN.\n");
  } while (OK==0);

  printf("Number of subintervals N = %d\n", N);
  printf("Interval number h = %lg\n", h);

  /* u[1][] is the particular solution - y1() */
  /* u[2][] is the derivative of particular solution */
  /* v[1][] is the homogeneous solution - y2() */
  /* v[2][] is the derivative of homogeneous solution */

  u = dmatrix(1,2,0,N);
  v = dmatrix(1,2,0,N);		
  k = dmatrix(1,4,1,4);		/* for y1() using RK4 method */
  K = dmatrix(1,4,1,4);		/* for y2() using RK4 method */

  /*-------------------------------------------------------------------------*/
				/* print out table header */
  printf("\n x\t u1\t\t v1\t\t w2 = y'(x)\t w1 = y(x)");
  printf("\n-----------------------------------");
  printf("-----------------------------------\n");
  /*-------------------------------------------------------------------------*/

  /* STEP #1 */
  u[1][0] = alpha;		/* y1(a) = alpha */
  u[2][0] = 0.0;		/* y1'(a) = 0 */
  v[1][0] = 0.0;		/* y2(a) = 0 */
  v[2][0] = 1.0;		/* y2'(a) = 1 */

  /* STEP #2 */
  for (i=0;i<N;i++) {		/* Do Steps 3-4. */

    /* STEP #3 */
    x = a + i*h;

    /* STEP #4 */
    k[1][1] = h*u[2][i];
    k[1][2] = h*(p(x)*u[2][i] + q(x)*u[1][i] + r(x));
    k[2][1] = h*(u[2][i] + 0.5*k[1][2]);
    k[2][2] = h*(p(x+h/2.0)*(u[2][i] + 0.5*k[1][2])
		 + q(x+h/2.0)*(u[1][i] + 0.5*k[1][1]) + r(x+h/2.0));
    k[3][1] = h*(u[2][i] + 0.5*k[2][2]);
    k[3][2] = h*(p(x+h/2.0)*(u[2][i] + 0.5*k[2][2])
		 + q(x+h/2.0)*(u[1][i] + 0.5*k[2][1]) + r(x+h/2.0));
    k[4][1] = h*(u[2][i] + k[3][2]);
    k[4][2] = h*(p(x+h)*(u[2][i]+k[3][2]) + q(x+h)*(u[1][i]+k[3][1]) + r(x+h));
    u[1][i+1] = u[1][i] + (k[1][1] + 2.0*k[2][1] + 2.0*k[3][1] + k[4][1])/6.0;
    u[2][i+1] = u[2][i] + (k[1][2] + 2.0*k[2][2] + 2.0*k[3][2] + k[4][2])/6.0;

    K[1][1] = h*v[2][i];
    K[1][2] = h*(p(x)*v[2][i] + q(x)*v[1][i]);
    K[2][1] = h*(v[2][i] + 0.5*K[1][2]);
    K[2][2] = h*(p(x+h/2.0)*(v[2][i] + 0.5*K[1][2])
		 + q(x+h/2.0)*(v[1][i] + 0.5*K[1][1]));
    K[3][1] = h*(v[2][i] + 0.5*K[2][2]);
    K[3][2] = h*(p(x+h/2.0)*(v[2][i] + 0.5*K[2][2])
		 + q(x+h/2.0)*(v[1][i] + 0.5*K[2][1]));
    K[4][1] = h*(v[2][i] + K[3][2]);
    K[4][2] = h*(p(x+h)*(v[2][i]+K[3][2]) + q(x+h)*(v[1][i]+K[3][1]));
    v[1][i+1] = v[1][i] + (K[1][1] + 2.0*K[2][1] + 2.0*K[3][1] + K[4][1])/6.0;
    v[2][i+1] = v[2][i] + (K[1][2] + 2.0*K[2][2] + 2.0*K[3][2] + K[4][2])/6.0;
  }

  /* STEP #5 */
  w10 = alpha;
  w20 = (beta - u[1][N]) / v[1][N];
  printf("% .2lf\t% .9lf\t% .9lf\t% .9lf\t% .9lf\n",
		 a, u[1][0], v[1][0], w20, w10);

  /* STEP #6 */
  for (i=1;i<=N;i++) {
    W1 = u[1][i] + w20 * v[1][i];
    W2 = u[2][i] + w20 * v[2][i];
    x = a + i*h;
    printf("% .2lf\t% .9lf\t% .9lf\t% .9lf\t% .9lf\n",
		 x, u[1][i], v[1][i], W2, W1);
  }

}  /* STEP #7 */		/* STOP - Process is complete. */

/*****************************************************************************/
/*****************************************************************************/
double p(x)		/* Equation p(x). */
double x;
{
  return (-2.0/x);
}

/*****************************************************************************/
double q(x)		/* Equation q(x). */
double x;
{
  return (2.0/(x*x));
}

/*****************************************************************************/
double r(x)		/* Equation r(x). */
double x;
{
  return (sin(log(x))/(x*x));
}
/*****************************************************************************/
