/******************************************************************************
		Poisson Equation Finite-Difference - Algorithm 12.1
*******************************************************************************

To approximate the solution of the Poisson equation
	d^2u(x,t)/dx^2 + d^2u(x,t)/dy^2 = f(x,y),  a <= x <= b, c <= y <= d,
subject to the boundry conditions
	u(x,y) = g(x,y)		if x = a or x = b and c <= y <= d
and
	u(x,y) = g(x,y)		if y = c or y = d and a <= x <= b:

INPUT endpoints a, b, c, d; integers m, n; tolerance TOL; maximum number of
iterations LBOUND; the functions u(), and f().

OUTPUT approximations w(i,j) to u(x(i),y(j)) for each i = 1,...,n-1 and j =
1,...,m-1 or a message that the maximum number of iterations was exceeded.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include <math.h>		/* for fabs() */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097

#define LBOUND	100		/* Maximum number of iterations. */


main()
{
  double *x, *y, **w, u(), f();
  double a, b, c, d, h, k, z, lambda, mu, NORM, TOL, temp;
  int i, j, l, m, n;

  get_title();

  printf("Poisson Equation Finite-Difference - Algorithm 12.1\n\n");
  fprintf(stderr, "Poisson Equation Finite-Difference - Algorithm 12.1\n\n");

  printf("d^2u(x,t)/dx^2 + d^2u(x,t)/dy^2 = f(x,y)\n");
  printf("f(x,y) = x*exp(y)\n");
  printf("u(0,y) = 0, u(2,y) = 2*exp(y)\n");
  printf("u(x,0) = x, u(x,1) = exp(1)*x\n\n");
  /* Needs updating    $   $   $   $   $  */

  fprintf(stderr, "Enter endpoint a: ");
  scanf("%lf", &a);
  fprintf(stderr, "Enter endpoint b: ");
  scanf("%lf", &b);
  fprintf(stderr, "Enter endpoint c: ");
  scanf("%lf", &c);
  fprintf(stderr, "Enter endpoint d: ");
  scanf("%lf", &d);
  printf("For %lg <= x <= %lg and %lg <= y <= %lg.\n", a, b, c, d);

  fprintf(stderr, "Enter n, the number of subintervals on x: ");
  scanf("%d", &n);
  if (n <= 0)
    naaerror("n must be > 0.");	/* ERROR. */
  printf("Number of subintervals on x is n = %d\n", n);

  fprintf(stderr, "Enter m, the number of subintervals on y: ");
  scanf("%d", &m);
  if (m <= 0)
    naaerror("m must be > 0.");	/* ERROR. */
  printf("Number of subintervals on y is m = %d\n", m);

  fprintf(stderr, "Enter tolerance, TOL: ");
  scanf("%lf", &TOL);
  printf("Tolerance: %lG\n\n\n", TOL);

  /*-------------------------------------------------------------------------*/
  printf(" i\t j\t    x(i)\t    y(i)\t w(i,j)");
  printf("\n------------------------------");
  printf("-------------------------------\n");
  /*-------------------------------------------------------------------------*/

  x=dvector(1,n-1);		/* For values on x-axis. */
  y=dvector(1,m-1);		/* For values on y-axis. */
  w=dmatrix(1,n-1,1,m-1);	/* For w(i,j). */

  /* STEP #1 */
  h = (b-a)/n;
  k = (d-c)/m;

  /* STEP #2 */
  for (i=1;i<n;i++)		/* Steps 2 and 3 construct mesh points. */
    x[i] = a + i*h;

  /* STEP #3 */
  for (j=1;j<m;j++)		/* Construct mesh points. */
    y[j] = c + j*k;

  /* STEP #4 */
  for (i=1;i<n;i++)
    for (j=1;j<m;j++)
      w[i][j] = 0.0;

  /* STEP #5 */
  lambda = (h*h) / (k*k);
  mu = 2.0*(1.0+lambda);

  /* STEP #6 and #20 */	/* Steps 7-20 perform Gauss-Seidel iterations. */
  for (l=1;l<=LBOUND;l++) {	/* Do Steps 7-20. */

    /* STEP #7 */
    z = (-h*h*f(x[1],y[m-1]) + u('a',a,y[m-1]) + lambda*u('d',x[1],d) +
		lambda*w[1][m-2] + w[2][m-1]) / mu;
    NORM = fabs(z-w[1][m-1]);
    w[1][m-1] = z;

    /* STEP #8 */
    for (i=2;i<=n-2;i++) {
      z = (-h*h*f(x[i],y[m-1]) + lambda*u('d',x[i],d) + w[i-1][m-1] +
		w[i+1][m-1] + lambda*w[i][m-2]) / mu;
      temp = fabs(w[i][m-1]-z);
      if (temp > NORM)
        NORM = temp;
      w[i][m-1] = z;
    }

    /* STEP #9 */
    z = (-h*h*f(x[n-1],y[m-1]) + u('b',b,y[m-1]) + lambda*u('d',x[n-1],d) +
		w[n-2][m-1] + lambda*w[n-1][m-2]) / mu;
    temp = fabs(w[n-1][m-1]-z);
    if (temp > NORM)
      NORM = temp;
    w[n-1][m-1] = z;

    /* STEP #10 */
    for (j=m-2;j>=2;j--) {	/* Do Steps 11-13. */

      /* STEP #11 */
      z = (-h*h*f(x[1],y[j]) + u('a',a,y[j]) + lambda*w[1][j+1] +
		lambda*w[1][j-1] + w[2][j]) / mu;
      temp = fabs(w[1][j]-z);
      if (temp > NORM)
        NORM = temp;
      w[1][j] = z;

      /* STEP #12 */
      for (i=2;i<=n-2;i++) {
        z = (-h*h*f(x[i],y[j]) + w[i-1][j] + lambda*w[i][j+1] + w[i+1][j] +
		lambda*w[i][j-1]) / mu;
        temp = fabs(w[i][j]-z);
        if (temp > NORM)
          NORM = temp;
        w[i][j] = z;
      }

      /* STEP #13 */
      z = (-h*h*f(x[n-1],y[j]) + u('b',b,y[j]) + w[n-2][j] +
		lambda*w[n-1][j+1] + lambda*w[n-1][j-1]) / mu;
      temp = fabs(w[n-1][j]-z);
      if (temp > NORM)
        NORM = temp;
      w[n-1][j] = z;
    }

    /* STEP #14 */
    z = (-h*h*f(x[1],y[1]) + u('a',a,y[1]) + lambda*u('c',x[1],c) +
		lambda*w[1][2] + w[2][1]) / mu;
    temp = fabs(w[1][1]-z);
    if (temp > NORM)
      NORM = temp;
    w[1][1] = z;

    /* STEP #15 */
    for (i=2;i<=n-2;i++) {
      z = (-h*h*f(x[i],y[1]) + lambda*u('c',x[i],c) + w[i-1][1] +
		lambda*w[i][2] + w[i+1][1]) / mu;
      temp = fabs(w[i][1]-z);
      if (temp > NORM)
        NORM = temp;
      w[i][1] = z;
    }

    /* STEP #16 */
    z = (-h*h*f(x[n-1],y[1]) + u('b',b,y[1]) + lambda*u('c',x[n-1],c) +
		w[n-2][1] + lambda*w[n-1][2]) / mu;
    temp = fabs(w[n-1][1]-z);
    if (temp > NORM)
      NORM = temp;
    w[n-1][1] = z;

    /* STEP #17 */
    if (NORM <= TOL) {		/* Do Steps 18-19. */

      /* STEP #18 */
      for (i=1;i<n;i++)
        for (j=1;j<m;j++)
          printf(" %d\t %d\t% 8.8lG\t% 8.8lG\t% 8.8lG\n",
		 i, j, x[i], y[j], w[i][j]);

      /* STEP #19 */
      printf("\nRequired %d iterations.\n", l);
      return (1);		/* STOP - Procedure completed successfully. */
    }
  }

  /* STEP #21 */
  printf("\nMaximum number of iterations (%d) exceeded.\n", LBOUND);

}				/* STOP */

/*****************************************************************************/
double f(x,y)		/* Function to evaluate. */
double x, y;
{
  return (x*exp(y));
}

/*****************************************************************************/
double u(ch,x,y)	/* Boundary conditions. */
char ch;
double x,y;
{
  switch (ch) {
    case 'a':  return (0.0);		/* For u(a,y). */
    case 'b':  return (2.0*exp(y));	/* For u(b,y). */
    case 'c':  return (x);		/* For u(x,c). */
    case 'd':  return (exp(1.0)*x);	/* For u(x,d). */
    default:   naaerror("Invalid U(x,y) function call.");
  }
}
/*****************************************************************************/
